import os
import platform


# ============================================
#                   get_os
# ============================================
def get_os() -> str:
    """
    Returns the operating system and "bitness" (64 or 32 bit):

        * windows_32bit
        * windows_64bit
        * pi_32bit
        * pi_64bit
        * linux_32bit
        * linux_64bit

    Returns
    -------
    os : str
        The name and "bitness" of the current operating system.
    """
    system = platform.system().lower()

    if system == "linux":
        machine = os.uname().machine
        if machine.startswith("arm") or machine.startswith("aarch"):
            system = "pi"

    return system + "_" + platform.architecture()[0]



from functools import wraps
from typing import Any, Callable

from botocore.exceptions import ClientError
from semantic_version import Version


# ============================================
#              requires_status
# ============================================
def requires_status(status) -> Callable:
    """
    Ensures we are either connected or streaming.
    """

    def status_decorator(func: Callable) -> Callable:
        @wraps(func)
        def status_wrapper(*args, **kwargs) -> Any:
            if not getattr(args[0], status):
                raise RuntimeError(f"Error: not {status}.")
            return func(*args, **kwargs)

        return status_wrapper

    return status_decorator


# ============================================
#                  validate
# ============================================
def validate(func) -> Callable:
    """
    Checks if the result of a command is SUCCESS.
    """

    @wraps(func)
    def validate_wrapper(*args, **kwargs) -> Any:
        retCode = func(*args, **kwargs)
        # pylint: disable-next=protected-access
        if retCode != args[0]._SUCCESS.value:
            raise RuntimeError(f"Command: {func.__name__} failed.")
        return retCode

    return validate_wrapper


# ============================================
#           minimum_required_version
# ============================================
def minimum_required_version(version: str) -> Callable:
    """
    Makes sure that the device's firmware is at least the given
    version.
    """

    def min_ver_decorator(func: Callable) -> Callable:
        @wraps(func)
        def min_ver_wrapper(*args, **kwargs) -> Any:
            try:
                deviceVersion = args[0].firmwareVersion
            except AttributeError as err:
                raise RuntimeError("Must decorate a device method.") from err

            try:
                assert deviceVersion >= Version(version)
            except AssertionError as err:
                msg = f"Cannot use: {func.__name__}, firmware too low."
                raise RuntimeError(msg) from err
            return func(*args, **kwargs)

        return min_ver_wrapper

    return min_ver_decorator


# ============================================
#             check_status_code
# ============================================
def check_status_code(func: Callable) -> Callable:
    @wraps(func)
    def check_status_wrapper(*args, **kwargs) -> Any:
        try:
            return func(*args, **kwargs)
        # boto3 raises a client error when we either try to download something
        # that doesn't exist, download something we don't have permission for,
        # or try to list the contents of a bucket when we don't have the
        # credentials to do so. These errors are differentiated by the
        # HTTPStatusCode key in the response
        except ClientError as err:
            statusCode = err.response["ResponseMetadata"]["HTTPStatusCode"]
            if statusCode == 404:
                msg = "Error: requested object could not be found on S3. "
                msg += "Please check spelling and/or path."
            elif statusCode == 403:
                msg = "Error: S3 permission denied. Please check your credentials "
                msg += "in '~/.aws/credentials' and make sure you're passing the "
                msg += "correct profile to the function."
            else:
                msg = f"Error: received status code {statusCode} from S3."
            raise RuntimeError(msg) from err

    return check_status_wrapper



import ctypes as c
from pathlib import Path

import semantic_version as sem

# ============================================
#             Path Configuration
# ============================================

# All dephy-related files are stored here (libs, firmware, device
# specs, api specs, and bootloading tools)
dephyPath = Path.home().joinpath(".dephy").expanduser().absolute()

firmwareVersionCacheFile = dephyPath.joinpath("available_versions.yaml")

# libsDir is the name of the directory (mirrored on S3), whereas
# libsPath is the full path to that directory on the local file system
libsDir = "precompiled_c_libs"
libsPath = dephyPath.joinpath(libsDir)

# Each operating system potentially has its own library file name
libFiles = {
    "windows_64bit": "libfx_plan_stack.dll",
    "windows_32bit": "libfx_plan_stack.dll",
    "linux_64bit": "libfx_plan_stack.so",
    "linux_32bit": "libfx_plan_stack.so",
    "pi_64bit": "libfx_plan_stack.so",
    "pi_32bit": "libfx_plan_stack.so",
}

# legacyDeviceSpecsDir is the name of the directory (mirrored on S3), whereas
# legacyDeviceSpecsPath is the full path to that directory on the local file system
legacyDeviceSpecsDir = "legacy_device_specs"
legacyDeviceSpecsPath = dephyPath.joinpath(legacyDeviceSpecsDir)


# ============================================
#              S3 Configuration
# ============================================

# This bucket is public and holds all of the pre-compiled c libraries, device
# specs, api specs, and bootloader tools
dephyPublicFilesBucket = "dephy-public-files"


# ============================================
#           Version Configuration
# ============================================

# Firmware before this version had a very different communication protocol,
# method of reading data, did not contain device side information, the
# libraries did not have information about their own version, different device
# error codes, and a different way of obtaining the device name. We call
# devices with firmware prior to this cutoff legacy devices
legacyCutoff = sem.Version("10.0.0")


# ============================================
#                 Constants
# ============================================
baudRate = 230400
minHeartbeat = 50


# ============================================
#                 Controllers
# ============================================

# These values are used by the fxSendMotorCommand function to tell the firmware
# what quantity we are trying to command
controllers = {
    "position": c.c_int(0),
    "voltage": c.c_int(1),
    "current": c.c_int(2),
    "impedance": c.c_int(3),
    "none": c.c_int(4),
}


# ============================================
#                 Bootloader
# ============================================

# These values are used to tell the firmware which chip we're trying to flash
bootloaderTargets = {
    "habs": 0,
    "re": 1,
    "ex": 2,
    "mn": 3,
    "bt": 4,
    "xbee": 5,
}


# ============================================
#                Error Codes
# ============================================

# These values encode the meaning of the various return values for the
# C functions being called from the library
deviceErrorCodes = {
    "UNDEFINED": c.c_int(0),
    "SUCCESS": c.c_int(1),
    "FAILURE": c.c_int(2),
    "INVALID_PARAM": c.c_int(3),
    "INVALID_DEVICE": c.c_int(4),
    "NOT_STREAMING": c.c_int(5),
}

legacyDeviceErrorCodes = {
    "SUCCESS": c.c_int(0),
    "FAILURE": c.c_int(1),
    "INVALID_PARAM": c.c_int(2),
    "INVALID_DEVICE": c.c_int(3),
    "NOT_STREAMING": c.c_int(4),
}


# ============================================
#                    Habs
# ============================================

# It's easier to list the devices that don't have habs
noHabs = [
    "actpack",
]


# ============================================
#                Device Names
# ============================================

# Only applies to legacy devices. These values encode the name of the device
INVALID_APP = c.c_int(-1)
ACTPACK = c.c_int(0)
EXO = c.c_int(1)  # XCs are reported as exos
MD = c.c_int(2)  # NOTE: This is only true for v9.1. For < 9.1, it's
# NetMaster, but those devices are not in the wild

deviceNames = {
    ACTPACK.value: "actpack",
    EXO.value: "exo",
    MD.value: "md",
}


# ============================================
#                   UTTs
# ============================================
# This isn't great, but getting the number of UTTs depends on a function that
# was introduced in version 10, but it turns out the UTT functions themselves
# were available before then. However, when trying to use the utt functions
# from version 9, we still need to know how many there are, so, since we're
# past v9, the number of UTTs is fixed, so we hard code it here, though it makes
# me sad to do so
nUttsV9 = 15



import ctypes as c
import os
from pathlib import Path
from typing import Tuple

from botocore.exceptions import EndpointConnectionError
from semantic_version import Version

import flexsea.utilities.constants as fxc

from .aws import s3_download
from .firmware import Firmware
from .system import get_os


# ============================================
#                get_c_library
# ============================================
def get_c_library(firmwareVersion: Version, libFile: Path | None) -> Tuple:
    """
    If we're given a library file to use, we make sure it exists. If
    we're not given a library file to use, we check for a cached file
    on disk corresponding to our firmware version. If we don't have one,
    we try to download it from S3. We then use ctypes to load the library.
    """
    if libFile is None:
        _os = get_os()
        libFile = fxc.libsPath.joinpath(str(firmwareVersion), _os, fxc.libFiles[_os])
        if not libFile.exists():
            libFile.parent.mkdir(parents=True, exist_ok=True)
            libObj = f"{fxc.libsDir}/{firmwareVersion}/{_os}/{libFile.name}"
            try:
                s3_download(libObj, fxc.dephyPublicFilesBucket, str(libFile), None)
            except EndpointConnectionError as err:
                msg = "Error: could not connect to the internet to download the "
                msg += "necessary C library file. Please connect to the internet and "
                msg += "try again."
                print(msg)
                raise err

    clib = _load_clib(libFile)

    return (_set_prototypes(clib, firmwareVersion), libFile)


# ============================================
#                _load_clib
# ============================================
def _load_clib(libFile: Path) -> c.CDLL:
    """
    Uses ctypes to actually create an interface to the library file. If
    we're on Windows, we have to additionally add several directories to
    the path.
    """
    if not libFile.is_absolute():
        libFile = libFile.expanduser().absolute()
    libFile = str(libFile)

    if "win" in get_os():
        try:
            for extraPath in os.environ["PATH"].split(";"):
                if os.path.exists(extraPath) and "mingw" in extraPath:
                    os.add_dll_directory(extraPath)
            os.add_dll_directory(libFile)
        except OSError as err:
            msg = f"Error loading precompiled library: `{libFile}`\n"
            msg += "The most likely cause is a mismatch between the Python, pip and "
            msg += "shell architectures.\nPlease ensure all three are either 32 or 64 "
            msg += "bit.\nKeep different versions isolated by virtual environments.\n"
            print(msg)
            raise err

    return c.cdll.LoadLibrary(libFile)


# ============================================
#               _set_prototypes
# ============================================
def _set_prototypes(clib: c.CDLL, firmwareVersion: Version) -> c.CDLL:
    # pylint: disable=too-many-statements
    # Open
    clib.fxOpen.argtypes = [c.c_char_p, c.c_uint, c.c_uint]
    clib.fxOpen.restype = c.c_int

    # Close
    clib.fxClose.argtypes = [
        c.c_uint,
    ]
    clib.fxClose.restype = c.c_uint

    # Start streaming
    clib.fxStartStreaming.argtypes = [c.c_uint, c.c_uint, c.c_bool]
    clib.fxStartStreaming.restype = c.c_int

    # Start streaming with safety
    if firmwareVersion >= Version("9.1.2"):
        clib.fxStartStreamingWithSafety.argtypes = [
            c.c_uint,
            c.c_uint,
            c.c_bool,
            c.c_uint16,
        ]
        clib.fxStartStreamingWithSafety.restype = c.c_int

    # Stop streaming
    clib.fxStopStreaming.argtypes = [
        c.c_uint,
    ]
    clib.fxStopStreaming.restype = c.c_int

    # Set gains
    clib.fxSetGains.argtypes = [
        c.c_uint,
        c.c_uint,
        c.c_uint,
        c.c_uint,
        c.c_uint,
        c.c_uint,
        c.c_uint,
    ]
    clib.fxSetGains.restype = c.c_int
    # Send motor command
    clib.fxSendMotorCommand.argtypes = [c.c_uint, c.c_int, c.c_int]
    clib.fxSendMotorCommand.restype = c.c_int

    # Find poles
    clib.fxFindPoles.argtypes = [
        c.c_uint,
    ]
    clib.fxFindPoles.restype = c.c_int

    # Activate bootloader
    clib.fxActivateBootloader.argtypes = [c.c_uint, c.c_uint8]
    clib.fxActivateBootloader.restype = c.c_int

    # Is bootloader active
    clib.fxIsBootloaderActivated.argtypes = [
        c.c_uint,
    ]
    clib.fxIsBootloaderActivated.restype = c.c_int

    # Request firmware version
    clib.fxRequestFirmwareVersion.argtypes = [
        c.c_uint,
    ]
    clib.fxRequestFirmwareVersion.restype = c.c_int

    # Get last received firmware version
    clib.fxGetLastReceivedFirmwareVersion.argtypes = [
        c.c_uint,
    ]
    clib.fxGetLastReceivedFirmwareVersion.restype = Firmware

    # Get device type value
    if firmwareVersion < Version("10.0.0"):
        clib.fxGetAppType.argtypes = [
            c.c_uint,
        ]
        clib.fxGetAppType.restype = c.c_int

    # Get read data queue size
    clib.fxGetReadDataQueueSize.argtypes = [
        c.c_uint,
    ]
    clib.fxGetReadDataQueueSize.restype = c.c_int

    if firmwareVersion >= Version("10.0.0"):
        # Get max device name length
        clib.fxGetMaxDeviceNameLength.argtypes = []
        clib.fxGetMaxDeviceNameLength.restype = c.c_int

        # Get device name
        clib.fxGetDeviceTypeNameWrapper.argtypes = [c.c_uint, c.c_char_p]
        clib.fxGetDeviceTypeNameWrapper.restype = c.c_int

        # Get max device side length
        clib.fxGetMaxDeviceSideNameLength.argtypes = []
        clib.fxGetMaxDeviceSideNameLength.restype = c.c_int

        # Get side
        clib.fxGetDeviceSideNameWrapper.argtypes = [c.c_uint, c.c_char_p]
        clib.fxGetDeviceSideNameWrapper.restype = c.c_int

        # Get libs version
        clib.fxGetLibsVersion.argtypes = [
            c.POINTER(c.c_uint16),
            c.POINTER(c.c_uint16),
            c.POINTER(c.c_uint16),
        ]
        clib.fxGetLibsVersion.restype = c.c_int

        # Get max field name length
        clib.fxGetMaxDataLabelLength.argtypes = []
        clib.fxGetMaxDataLabelLength.restype = c.c_int

        # Get fields
        clib.fxGetDataLabelsWrapper.argtypes = [
            c.c_uint,
            c.POINTER(c.POINTER(c.c_char)),
            c.POINTER(c.c_int),
        ]
        clib.fxGetDataLabelsWrapper.restype = c.c_int

        # Get max data elements
        clib.fxGetMaxDataElements.argtypes = []
        clib.fxGetMaxDataElements.restype = c.c_int

        # Request uvlo
        clib.fxRequestUVLO.argtypes = [
            c.c_uint,
        ]
        clib.fxRequestUVLO.restype = c.c_int

        # Read uvlo
        clib.fxGetLastReceivedUVLO.argtypes = [
            c.c_uint,
        ]
        clib.fxGetLastReceivedUVLO.restype = c.c_int

        # Set uvlo
        clib.fxSetUVLO.argtypes = [c.c_uint, c.c_uint]
        clib.fxSetUVLO.restype = c.c_int

        # Get num utts
        clib.fxGetNumUtts.argtypes = []
        clib.fxGetNumUtts.restype = c.c_int

        # Set utts
        clib.fxSetUTT.argtypes = [c.c_uint, c.POINTER(c.c_int), c.c_uint, c.c_byte]
        clib.fxSetUTT.restype = c.c_int

        # Reset utts
        clib.fxSetUTTsToDefault.argtypes = [
            c.c_uint,
        ]
        clib.fxSetUTTsToDefault.restype = c.c_int

        # Save utts
        clib.fxSaveUTTToMemory.argtypes = [
            c.c_uint,
        ]
        clib.fxSaveUTTToMemory.restype = c.c_int

        # Request utts
        clib.fxRequestUTT.argtypes = [
            c.c_uint,
        ]
        clib.fxRequestUTT.restype = c.c_int

        # Get last received utts
        clib.fxGetLastReceivedUTT.argtypes = [c.c_uint, c.POINTER(c.c_int), c.c_uint]
        clib.fxGetLastReceivedUTT.restype = c.c_int

        # IMU Calibration
        clib.fxSetImuCalibration.argtypes = [
            c.c_uint,
        ]
        clib.fxSetImuCalibration.restype = c.c_int

    return clib


# ============================================
#             set_read_functions
# ============================================
def set_read_functions(
    clib: c.CDLL, deviceName: str, isLegacy: bool, deviceType: c.Structure | None
) -> c.CDLL:
    """
    Sets the prototypes for the read and read_all functions. This
    is done here and not with the rest of the prototypes because,
    for legacy devices, we need the device name, which we can't get
    until we call open, for which we need the function prototypes...
    We do it here for non-legacy devices because it's easier than
    passing and worrying about the firmware version to set_prototypes
    """
    if isLegacy:
        if deviceName == "actpack":
            readFunc = getattr(clib, "fxReadDevice")
            readAllFunc = getattr(clib, "fxReadDeviceAll")
        # This also covers XCs, since thoes are reported as exos
        elif deviceName == "exo":
            readFunc = getattr(clib, "fxReadExoDevice")
            readAllFunc = getattr(clib, "fxReadExoDeviceAll")
        elif deviceName == "md":
            readFunc = getattr(clib, "fxReadMdDevice")
            readAllFunc = getattr(clib, "fxReadMdDeviceAll")
        else:
            raise ValueError(f"Unknown device: {deviceName}")

        readFunc.argtypes = [c.c_uint, c.POINTER(deviceType)]
        readFunc.restype = c.c_int
        setattr(clib, "read", readFunc)

        readAllFunc.argtypes = [c.c_uint, c.POINTER(deviceType), c.c_uint]
        readAllFunc.restype = c.c_int
        setattr(clib, "read_all", readAllFunc)

    else:
        readFunc = getattr(clib, "fxReadDevice")
        readFunc.argtypes = [c.c_uint, c.POINTER(c.c_int32), c.POINTER(c.c_int)]
        readFunc.restype = c.c_int
        setattr(clib, "read", readFunc)

        readAllFunc = getattr(clib, "fxReadDeviceAllWrapper")
        readAllFunc.argtypes = [
            c.c_uint,
            c.POINTER(c.POINTER(c.c_int32)),
            c.POINTER(c.c_int),
        ]
        readAllFunc.restype = None
        setattr(clib, "read_all", readAllFunc)

    return clib



import hashlib
from pathlib import Path
from typing import List

import boto3
from botocore import UNSIGNED
from botocore.client import BaseClient, Config
from botocore.exceptions import ClientError
from botocore.exceptions import ProfileNotFound

from flexsea.utilities.decorators import check_status_code


# ============================================
#                 s3_download
# ============================================
@check_status_code
def s3_download(obj: str, bucket: str, dest: str, profile: str | None = None) -> None:
    """
    Downloads `obj` from `bucket` to `dest` with the AWS
    credentials profile `profile`.
    """
    # https://stackoverflow.com/a/34866092
    if profile is None:
        client = boto3.client(
            "s3", config=Config(signature_version=UNSIGNED), region_name="us-east-1"
        )
    else:
        try:
            session = boto3.Session(profile_name=profile)
        except ProfileNotFound as err:
            msg = f"Error: invalid AWS profile `{profile}`"
            raise ValueError(msg) from err
        client = session.client("s3")

    try:
        client.download_file(bucket, obj, dest)
    except ClientError:
        # If the download fails, one possible reason is because we weren't given a
        # valid object path, but, instead, just a base name, e.g., myfirmware.dfu
        # instead of firmwareBucket/major.minor.patch/device/hw/myfirmware.dfu
        # In this case we want to search the given bucket for the file
        obj = s3_find_object(obj, bucket, client)
        client.download_file(bucket, obj, dest)
    _validate_download(client, bucket, obj, dest)
    print(f"Downloaded {obj} from {bucket} to {dest}")


# ============================================
#              s3_find_object
# ============================================
@check_status_code
def s3_find_object(fileName: str, bucket: str, client: str) -> str:
    """
    Searches the given bucket for the given file. Returns the full object
    path if there's only one match. If there aren't any matches or there's
    more than one, we fail.
    """
    # https://tinyurl.com/4scnuk6c
    paginator = client.get_paginator("list_objects_v2")
    pageIterator = paginator.paginate(Bucket=bucket)
    objects = pageIterator.search(f"Contents[?contains(Key, `{fileName}`)][]")
    # There should only be one match
    # Objects is a generator, so we have to convert it to a list to check length
    items = []
    for item in objects:
        items.append(item["Key"])
    if len(items) == 0:
        raise FileNotFoundError(f"Could not find: {fileName} in {bucket}")
    if len(items) > 1:
        raise FileNotFoundError(f"Found multiple options for: {fileName} in {bucket}")
    return items[0]


# ============================================
#             _validate_download
# ============================================
def _validate_download(
    client: BaseClient, bucket: str, fileObj: str, dest: str
) -> None:
    """
    Compares the AWS md5 hash to the local md5 hash to make sure the
    files are the same.

    S3 objects have an attribute called 'ETag', which is a string.
    There are two possibilities: the string is a hex digest or the
    string is a hex digest + -NUMBER.

    In the case that ETag has no -NUMBER suffix, it means that the ETag
    is the md5 hash of the file contents.

    In the case that there is a -NUMBER suffix, it means that the file
    was uploaded to S3 in NUMBER chunks and that the hex digest is actually
    the hex digest of the digests of all of the chunks concatenated together.

    Parameters
    ----------
    client : BaseClient
        The object that allows use to communicate with S3.

    bucket : str
        The name of the bucket the file came from.

    fileObj : str
        The name of the fileObj we are validating.

    dest : str
        The name of the dowloaded file on disk.

    Raises
    -------
    FileNotFoundError
        If the downloaded file is not present in the desired destinationself.

    AssertionError
        If the hash of the remote object does not match the hash of the local
        file.
    """
    try:
        assert Path(dest).exists()
    except AssertionError as err:
        raise FileNotFoundError from err

    # Check the local file's integrity by comparing its md5 hash to
    # AWS's md5 hash, called ETag
    objData = client.head_object(Bucket=bucket, Key=fileObj)
    etag = objData["ETag"].strip('"')

    try:
        nChunks = int(etag.split("-")[1])
    except IndexError:
        nChunks = 1

    if nChunks == 1:
        with open(dest, "rb") as fd:
            data = fd.read()
        localHash = hashlib.md5(data).hexdigest()

    elif nChunks > 1:
        chunkHashes = []
        with open(dest, "rb") as fd:
            for chunk in range(1, nChunks + 1):
                objData = client.head_object(
                    Bucket=bucket, Key=fileObj, PartNumber=chunk
                )
                chunkSize = objData["ContentLength"]
                data = fd.read(chunkSize)
                if data:
                    chunkHashes.append(hashlib.md5(data))
                else:
                    break

        if len(chunkHashes) == 1:
            localHash = chunkHashes[0].hexdigest()
        else:
            digests = b"".join([m.digest() for m in chunkHashes])
            digestsMd5 = hashlib.md5(digests)
            localHash = f"{digestsMd5.hexdigest()}-{len(chunkHashes)}"

    assert localHash == etag


# ============================================
#                get_s3_objects
# ============================================
@check_status_code
def get_s3_objects(bucket: str, client: BaseClient, prefix: str = "") -> List:
    """
    Recursively loops over all directories in a bucket and returns a
    list of files.

    Parameters
    ----------
    bucket : str
        The name of the bucket we're getting files from.

    client : botocore.client.BaseClient
        The object providing an interface to S3.

    prefix : str
        The directory we're looping over. If `""`, then we get the
        top-level directories.

    Returns
    -------
    List[str]
        A list of all the objects in the bucket.
    """
    objectList = []
    objects = client.list_objects_v2(Bucket=bucket, Delimiter="/", Prefix=prefix)

    if "CommonPrefixes" in objects:
        for pre in objects["CommonPrefixes"]:
            objectList += get_s3_objects(bucket, client, pre["Prefix"])

    if "Contents" in objects:
        return objectList + [obj["Key"] for obj in objects["Contents"][1:]]
    return objectList



import ctypes as c
from pathlib import Path
import sys
from time import sleep
from typing import List

from semantic_version import Version

import flexsea.utilities.constants as fxc
from flexsea.utilities.decorators import minimum_required_version
from flexsea.utilities.decorators import requires_status
from flexsea.utilities.decorators import validate
from flexsea.utilities.firmware import decode_firmware
from flexsea.utilities.firmware import validate_given_firmware_version
from flexsea.utilities.library import get_c_library
from flexsea.utilities.library import set_read_functions
from flexsea.utilities.specs import get_device_spec


# ============================================
#                    Device
# ============================================
class Device:
    """
    Representation of one of Dephy's devices. Serves as a way to
    send commands to -- and read data from -- a Dephy device.
    """

    # -----
    # constructor
    # -----
    def __init__(
        self,
        firmwareVersion: str,
        port: str,
        baudRate: int = fxc.baudRate,
        libFile: str = "",
        logLevel: int = 4,
        interactive: bool = True,
        debug: bool = False,
    ) -> None:
        if not debug:
            sys.tracebacklimit = 0
        # These are first so the destructor won't complain about the
        # class not having connected and streaming attributes if getting
        # and loading the C library fails
        self.connected: bool = False
        self.streaming: bool = False

        self.port: str = port
        self.interactive = interactive

        self.firmwareVersion = validate_given_firmware_version(
            firmwareVersion, self.interactive
        )

        if libFile:
            self.libFile = Path(libFile).expanduser().absolute()
            if not self.libFile.is_file():
                raise FileNotFoundError(f"Could not find library: {self.libFile}")
        else:
            self.libFile = None

        if baudRate <= 0:
            raise ValueError("Error: baud rate must be positive.")
        self.baudRate = baudRate

        if logLevel < 0 or logLevel > 6:
            raise ValueError("Log level must be in [0, 6].")
        self.logLevel = logLevel

        self.heartbeat: int = 0
        self.id: int = 0
        self.streamingFrequency: int = 0

        (self._clib, self.libFile) = get_c_library(self.firmwareVersion, self.libFile)

        self._fields: List[str] | None = None
        self._gains: dict = {}
        self._hasHabs: bool | None = None
        self._name: str = ""
        self._side: str = ""
        self._state: c.Structure | None = None
        self._stateType: c.Structure | None = None

        if self.firmwareVersion < fxc.legacyCutoff:
            self._SUCCESS = c.c_int(0)
            self._FAILURE = c.c_int(1)
            self._INVALID_PARAM = c.c_int(2)
            self._INVALID_DEVICE = c.c_int(3)
            self._NOT_STREAMING = c.c_int(4)
            self._isLegacy = True
            self._libVersion = "undefined"

        else:
            self._UNDEFINED = c.c_int(0)
            self._SUCCESS = c.c_int(1)
            self._FAILURE = c.c_int(2)
            self._INVALID_PARAM = c.c_int(3)
            self._INVALID_DEVICE = c.c_int(4)
            self._NOT_STREAMING = c.c_int(5)
            self._isLegacy = False
            self._libVersion = self._get_lib_version()

        print(f"Using firmware version: {self.firmwareVersion}")
        print(f"Using library file: {self.libFile}")

    # -----
    # destructor
    # -----
    def __del__(self) -> None:
        self.close()

    # -----
    # open
    # -----
    def open(self) -> None:
        """
        Establish a connection to a device. This is needed in order
        to send commands to the device and/or receive data from the
        device.
        """
        if self.connected:
            print("Already connected.")
            return

        port = self.port.encode("utf-8")

        self.id = self._clib.fxOpen(port, self.baudRate, self.logLevel)

        if self.id in (self._INVALID_DEVICE.value, -1):
            raise RuntimeError("Failed to connect to device.")

        self.connected = True

        self._name = self.name
        self._side = self.side
        self._hasHabs = self._name not in fxc.noHabs

        self._get_info_for_reading()

    # -----
    # _get_info_for_reading
    # -----
    def _get_info_for_reading(self) -> None:
        if self._isLegacy:
            self._get_state()
        else:
            self._get_fields()
        self._clib = set_read_functions(
            self._clib, self._name, self._isLegacy, self._stateType
        )

    # -----
    # _get_state
    # -----
    def _get_state(self) -> None:
        stateSpec = get_device_spec(self._name, self.firmwareVersion)

        class LegacyDeviceState(c.Structure):
            _pack_ = 1
            _fields_ = [(k, getattr(c, v)) for k, v in stateSpec.items()]

        self._stateType = LegacyDeviceState
        self._state = LegacyDeviceState()

    # -----
    # _get_fields
    # -----
    def _get_fields(self) -> None:
        maxFields = self._clib.fxGetMaxDataElements()
        maxFieldLength = self._clib.fxGetMaxDataLabelLength()
        nLabels = c.c_int()

        # https://code.activestate.com/lists/python-list/704158
        labels = (c.POINTER(c.c_char) * maxFields)()

        for i in range(maxFields):
            labels[i] = c.create_string_buffer(maxFieldLength)

        retCode = self._clib.fxGetDataLabelsWrapper(self.id, labels, c.byref(nLabels))

        if retCode != self._SUCCESS.value:
            raise RuntimeError("Could not get device field labels.")

        # Convert the labels from chars to python strings
        fields = [""] * nLabels.value
        if self._fields is None:
            self._fields = []
        for i in range(nLabels.value):
            for j in range(maxFieldLength):
                fields[i] += labels[i][j].decode("utf8")
            fields[i] = fields[i].strip("\x00")
            self._fields.append(fields[i])

    # -----
    # close
    # -----
    def close(self) -> None:
        """
        Severs connection with device. Will no longer be able to send
        commands or receive data.
        """
        if self.streaming:
            self.stop_streaming()

        if self.connected:
            self.stop_motor()
            self._clib.fxClose(self.id)
            self.connected = False

    # -----
    # start_streaming
    # -----
    @requires_status("connected")
    def start_streaming(
        self, frequency: int, heartbeat: int = 0, useSafety: bool = False
    ) -> None:
        """
        Instructs the device to send us data at the given `frequency`.

        Parameters
        ----------
        frequency : int
            The number of times per second the device should send data

        heartbeat : int
            When streaming, the computer periodically sends a message to
            the device to let it know that the connection between them
            is still alive. These are called heartbeat messages. This
            variable specifies the amount of time (in milliseconds)
            between successive heartbeat messages. This is related to
            how long the device will wait without receiving a heartbeat
            message before shutting itself off (five times `heartbeat`).

        useSafety : bool
            If `True`, the device will stop the motor if it doesn't receive
            a heartbeat message in time.

        Raises
        ------
        ValueError
            If `frequency` or `heartbeat` are invalid.
        """
        if self.streaming:
            print("Already streaming.")
            return

        try:
            assert frequency > 0
        except AssertionError as err:
            raise ValueError("Frequency must be > 0.") from err

        self.streamingFrequency = frequency

        self.heartbeat = heartbeat

        if useSafety:
            self._stream_with_safety()
        else:
            self._stream_without_safety()

        self.streaming = True

    # -----
    # _stream_with_safety
    # -----
    @minimum_required_version("9.1.0")
    @validate
    def _stream_with_safety(self) -> int:
        try:
            assert fxc.minHeartbeat <= self.heartbeat < self.streamingFrequency
        except AssertionError as err:
            msg = f"Heartbeat must be in [{fxc.minHeartbeat}, frequency]"
            raise ValueError(msg) from err

        return self._clib.fxStartStreamingWithSafety(
            self.id, self.streamingFrequency, 1, self.heartbeat
        )

    # -----
    # _stream_without_safety
    # -----
    @validate
    def _stream_without_safety(self) -> int:
        return self._clib.fxStartStreaming(self.id, self.streamingFrequency, 1)

    # -----
    # stop_streaming
    # -----
    @requires_status("streaming")
    def stop_streaming(self) -> int:
        """
        Instructs the device to stop sending data.
        """
        retCode = self._clib.fxStopStreaming(self.id)

        if retCode != self._SUCCESS.value:
            raise RuntimeError("Failed to stop streaming.")

        self.streaming = False

    # -----
    # set_gains
    # -----
    @requires_status("connected")
    def set_gains(self, kp: int, ki: int, kd: int, k: int, b: int, ff: int) -> None:
        """
        Sets the gains used by PID controllers on the device.

        Parameters
        ----------
        kp : int
            Proportional gain.

        ki : int
            Integral gain.

        kd : int
            Differential gain.

        k : int
            Stiffness (used in impedence control only).

        b : int
            Damping (used in impedance control only).

        ff : int
            Feed forward gain.
        """
        # There is a bug either on the C side or in the firmware where,
        # sometimes, the gains aren't set, so we try multiple times
        for _ in range(5):
            returnCode = self._clib.fxSetGains(self.id, kp, ki, kd, k, b, ff)
            sleep(0.001)
        if returnCode != self._SUCCESS.value:
            raise RuntimeError("Failed to set gains.")
        self._gains = {"kp": kp, "ki": ki, "kd": kd, "k": k, "b": b, "ff": ff}

    # -----
    # command_motor_position
    # -----
    @requires_status("connected")
    @validate
    def command_motor_position(self, value: int) -> int:
        """
        Sets motor to given position.

        Parameters
        ----------
        value : int
            Desired motor position in encoder units.
        """
        controller = fxc.controllers["position"]
        return self._clib.fxSendMotorCommand(self.id, controller, value)

    # -----
    # command_motor_current
    # -----
    @requires_status("connected")
    @validate
    def command_motor_current(self, value: int) -> int:
        """
        Sends the given current to the motor.

        Parameters
        ----------
        value : int
            Desired motor current in milli-Amps.
        """
        controller = fxc.controllers["current"]
        return self._clib.fxSendMotorCommand(self.id, controller, value)

    # -----
    # command_motor_voltage
    # -----
    @requires_status("connected")
    @validate
    def command_motor_voltage(self, value: int) -> int:
        """
        Sets motor's voltage.

        Parameters
        ----------
        value : int
            Desired motor voltage in milli-volts.
        """
        controller = fxc.controllers["voltage"]
        return self._clib.fxSendMotorCommand(self.id, controller, value)

    # -----
    # command_motor_impedance
    # -----
    @requires_status("connected")
    @validate
    def command_motor_impedance(self, value: int) -> int:
        """
        Has the motor simulate a stretched from current position to
        the desired position using the damping gain b and stiffness
        gain k.

        Parameters
        ----------
        value : int
            Desired motor position in ticks.
        """
        controller = fxc.controllers["impedance"]
        return self._clib.fxSendMotorCommand(self.id, controller, value)

    # -----
    # stop_motor
    # -----
    @requires_status("connected")
    def stop_motor(self) -> int:
        """
        Stops the motor and resets the gains to zero.
        """
        controller = fxc.controllers["none"]
        retCode = self._clib.fxSendMotorCommand(self.id, controller, 0)
        if retCode != self._SUCCESS.value:
            raise RuntimeError("Failed to stop motor.")
        self._gains = {"kp": 0, "ki": 0, "kd": 0, "k": 0, "b": 0, "ff": 0}
        return retCode

    # -----
    # activate_bootloader
    # -----
    @requires_status("connected")
    def activate_bootloader(self, target: str) -> None:
        """
        Activates the bootloader for `target`.

        Parameters
        ----------
        target : str
            Bootloader target. Can be: mn, ex, re, or habs.
        """
        # We deliberately aren't using @validate here because we need
        # access to the IOError in `set_tunnen_mode`
        targetCode = fxc.bootloaderTargets[target]
        retVal = self._clib.fxActivateBootloader(self.id, targetCode)

        if retVal == self._INVALID_DEVICE.value:
            raise RuntimeError(f"Invalid device ID for: `{target}`.")
        if retVal != self._SUCCESS.value:
            raise IOError

    # -----
    # bootloaderActive
    # -----
    @property
    @requires_status("connected")
    def bootloaderActive(self) -> bool:
        """
        Returns whether or not the bootloader is active or not.

        Returns
        -------
        bool
            `True` if bootloader is active and `False` otherwise.
        """
        returnCode = self._clib.fxIsBootloaderActivated(self.id)
        if returnCode != self._SUCCESS.value:
            return False
        return True

    # -----
    # set_tunnel_mode
    # -----
    @requires_status("connected")
    def set_tunnel_mode(self, target: str, timeout: int = 30) -> bool:
        """
        All communication goes through Manage, so we need to put it into
        tunnel mode in order to activate the other bootloaders. When bootloading
        Manage itself, this causes it to reboot in DFU mode.

        If bootloading fails for any reason once tunnel mode has been set,
        the device is bricked and must be programmed manually. The exception
        to this rule is Habs.

        Parameters
        ----------
        target : str
            The name of the target to set (abbreviated).

        timeout : int
            The number of seconds to wait for confirmation before failing.

        Returns
        -------
        activated : bool
            If `True`, the bootloader was set successfully. If `False` then
            something went wrong.
        """
        activated = False

        while timeout > 0 and not activated:
            if timeout % 5 == 0:
                try:
                    self.activate_bootloader(target)
                except IOError:
                    pass

            # Device gets disconnected briefly when Mn resets, so we wait
            sleep(1)
            timeout -= 1
            activated = self.bootloaderActive

        return activated

    # -----
    # firmware_version
    # -----
    @property
    @requires_status("connected")
    def firmware_version(self) -> dict:
        """
        Gets the fimware versions of device's MCUs.

        Returns
        -------
        Dict
            A dictionary with the semantic version strings of manage,
            execute, and regulate's firmware. And habs, if applicable.
        """
        self._clib.fxRequestFirmwareVersion(self.id)

        sleep(5)

        fw = self._clib.fxGetLastReceivedFirmwareVersion(self.id)

        fwDict = {
            "mn": decode_firmware(fw.mn),
            "ex": decode_firmware(fw.ex),
            "re": decode_firmware(fw.re),
        }

        if self._hasHabs:
            fwDict["habs"] = decode_firmware(fw.habs)

        return fwDict

    # -----
    # print
    # -----
    @requires_status("streaming")
    def print(self, data: dict | None = None) -> None:
        """
        Reads the data from the device and then prints it to the screen.
        If data is given, we print that instead.
        """
        if data is None:
            data = self.read()
        for key, value in data.items():
            print(f"{key} : {value}")

    # -----
    # read
    # -----
    @requires_status("streaming")
    def read(self, allData: bool = False) -> dict | List[dict]:
        if allData:
            if self._isLegacy:
                return self._read_all_legacy()
            return self._read_all()
        if self._isLegacy:
            return self._read_legacy()
        return self._read()

    # -----
    # _read_all_legacy
    # -----
    def _read_all_legacy(self) -> List[dict]:
        qs = self._clib.fxGetReadDataQueueSize(self.id)
        data = (self._stateType * qs)()
        allData = []

        nRead = self._clib.read_all(self.id, data, qs)

        for i in range(nRead):
            # pylint: disable-next=protected-access
            allData.append({f[0]: getattr(data[i], f[0]) for f in data[i]._fields_})

        return allData

    # -----
    # _read_all
    # -----
    def _read_all(self) -> List[dict]:
        qs = self._clib.fxGetReadDataQueueSize(self.id)
        data = (c.POINTER(c.c_int32) * qs)()
        nElements = c.c_int()

        for i in range(qs):
            data[i] = (c.c_int32 * len(self._fields))()

        self._clib.read_all(self.id, data, c.byref(nElements))

        try:
            assert nElements.value == len(self._fields)
        except AssertionError as err:
            print("Different number of fields read than expected.")
            raise err

        allData = []

        for i in range(qs):
            singleTimeStepData = []
            for j in range(nElements.value):
                singleTimeStepData.append(data[i][j])
            allData.append(dict(zip(self._fields, singleTimeStepData)))

        return allData

    # -----
    # _read_legacy
    # -----
    def _read_legacy(self) -> dict:
        if self._clib.read(self.id, c.byref(self._state)) != self._SUCCESS.value:
            raise RuntimeError("Error: read command failed.")
        # pylint: disable-next=protected-access
        return {f[0]: getattr(self._state, f[0]) for f in self._state._fields_}

    # -----
    # _read
    # -----
    def _read(self) -> dict:
        maxDataElements = self._clib.fxGetMaxDataElements()
        nFields = c.c_int()
        deviceData = (c.c_int32 * maxDataElements)()

        retCode = self._clib.read(self.id, deviceData, c.byref(nFields))

        if retCode != self._SUCCESS.value:
            raise RuntimeError("Could not read from device.")

        try:
            assert nFields.value == len(self._fields)
        except AssertionError as err:
            print("Incorrect number of fields read.")
            raise err

        data = [deviceData[i] for i in range(nFields.value)]

        return dict(zip(self._fields, data))

    # -----
    # find_poles
    # -----
    @requires_status("connected")
    @validate
    def find_poles(self) -> int:
        """
        Instructs the device to go through the pole-finding process to
        align the motor correctly.
        """
        if not self.interactive:
            userInput = input(
                "WARNING: You should not use this function unless you know what "
                "you are doing!\nProceed?[y/n] "
            )

            if userInput != "y":
                print("Aborting pole finding.")
                return self._FAILURE.value

        msg = "NOTE: Please wait for the process to complete. The motor will stop "
        msg += "moving when it is done. It usually takes 1-2 minutes to complete."
        msg += "NOTE: Once the pole-finding procedure is completed, you must "
        msg += "power cylce the device for the changes to take effect."
        print(msg)

        return self._clib.fxFindPoles(self.id)

    # -----
    # uvlo - getter
    # -----
    @property
    @requires_status("connected")
    def uvlo(self) -> int:
        """
        Gets the currently set UVLO.
        """
        self._clib.fxRequestUVLO(self.id)
        sleep(5)
        return self._clib.fxGetLastReceivedUVLO(self.id)

    # -----
    # uvlo - setter
    # -----
    @uvlo.setter
    @validate
    def uvlo(self, value: int) -> int:
        """
        Sets the UVLO value for the device. `value` needs to be in milli-volts.
        """
        return self._clib.fxSetUVLO(self.id, value)

    # -----
    # calibrate_imu
    # -----
    @requires_status("connected")
    @validate
    def calibrate_imu(self) -> int:
        """
        Instructs the device to go through the IMU calibration process.
        """
        if not self.interactive:
            userInput = input(
                "WARNING: You should not use this function unless you know what "
                "you are doing!\nProceed?[y/n] "
            )

            if userInput != "y":
                print("Aborting IMU calibration.")
                return self._FAILURE.value

        return self._clib.fxSetImuCalibration(self.id)

    # -----
    # hasHabs
    # -----
    @property
    @requires_status("connected")
    def hasHabs(self) -> bool:
        return self._hasHabs

    # -----
    # name
    # -----
    @property
    @requires_status("connected")
    def name(self) -> str:
        if self._name:
            return self._name
        if self._isLegacy:
            self._name = self._get_legacy_name()
        else:
            self._name = self._get_name()
        return self._name

    # -----
    # _get_name
    # -----
    def _get_name(self) -> str:
        maxDeviceNameLength = self._clib.fxGetMaxDeviceNameLength()
        deviceName = (c.c_char * maxDeviceNameLength)()

        retCode = self._clib.fxGetDeviceTypeNameWrapper(self.id, deviceName)
        if retCode != self._SUCCESS.value:
            raise RuntimeError("Could not get device name.")

        return deviceName.value.decode("utf8").lower()

    # -----
    # _get_legacy_name
    # -----
    def _get_legacy_name(self) -> str:
        deviceTypeCode = self._clib.fxGetAppType(self.id)

        try:
            devName = fxc.deviceNames[deviceTypeCode]
        except KeyError as err:
            raise RuntimeError("Could not get device name.") from err

        return devName

    # -----
    # side
    # -----
    @property
    @requires_status("connected")
    def side(self) -> str:
        if self._side:
            return self._side
        if self._isLegacy:
            return "undefined"
        maxDeviceSideLength = self._clib.fxGetMaxDeviceSideNameLength()
        deviceSide = (c.c_char * maxDeviceSideLength)()

        retCode = self._clib.fxGetDeviceSideNameWrapper(self.id, deviceSide)
        if retCode != self._SUCCESS.value:
            raise RuntimeError("Could not get device side.")

        side = deviceSide.value.decode("utf8")

        # If side isn't applicable (for, e.g., an actpack), string is empty
        self._side = side.lower() if side else "undefined"
        return self._side

    # -----
    # _get_lib_version
    # -----
    @minimum_required_version("10.0.0")
    def _get_lib_version(self) -> Version:
        major = c.c_uint16()
        minor = c.c_uint16()
        patch = c.c_uint16()

        returnCode = self._clib.fxGetLibsVersion(
            c.byref(major), c.byref(minor), c.byref(patch)
        )

        if returnCode != self._SUCCESS.value:
            raise RuntimeError("Could not determine c library version.")

        return Version(f"{major.value}.{minor.value}.{patch.value}")

    # -----
    # num_utts
    # -----
    @property
    @minimum_required_version("9.1.0")
    @requires_status("connected")
    def num_utts(self) -> int:
        """
        The number of available UTT values is saved as #define NUM_UTT_VALS
        in the C library, so we have this convenience wrapper to access it.
        """
        if self.firmwareVersion.major == 9:
            return fxc.nUttsV9
        return self._clib.fxGetNumUtts()

    # -----
    # set_all_utts
    # -----
    @minimum_required_version("9.1.0")
    @requires_status("connected")
    @validate
    def set_all_utts(self, uttVals: List[int]) -> int:
        """
        Takes in a list of integer values and assigns each one to a UTT
        value.

        Parameters
        ----------
        uttVals : List[int]
            List of values, one for each UTT
        """
        numUtts = self.num_utts
        nVals = len(uttVals)

        try:
            assert nVals <= numUtts
        except AssertionError as err:
            raise ValueError("Error: too many UTT values given.") from err

        data = (c.c_int * nVals)()

        for i in range(nVals):
            data[i] = c.c_int(uttVals[i])

        return self._clib.fxSetUTT(self.id, data, nVals, c.c_byte(-1))

    # -----
    # set_utt
    # -----
    @minimum_required_version("9.1.0")
    @requires_status("connected")
    @validate
    def set_utt(self, uttVal: int, index: int) -> int:
        """
        Sets the value of a single UTT.

        Parameters
        ----------
        uttVal : int
            The value the desired UTT will be set to

        index : int
            The array index for the desired UTT
        """
        numUtts = self.num_utts

        try:
            assert 0 <= index < numUtts
        except AssertionError:
            print("Error: invalid UTT index.")

        data = (c.c_int * numUtts)()

        data[index] = c.c_int(uttVal)

        return self._clib.fxSetUTT(self.id, data, numUtts, c.c_byte(index))

    # -----
    # reset_utts
    # -----
    @minimum_required_version("9.1.0")
    @requires_status("connected")
    @validate
    def reset_utts(self) -> int:
        """
        Resets all UTTs to their default values.
        """
        return self._clib.fxSetUTTsToDefault(self.id)

    # -----
    # save_utts
    # -----
    @minimum_required_version("9.1.0")
    @requires_status("connected")
    @validate
    def save_utts(self) -> int:
        """
        Saves the current UTT values to the device's internal memory so
        that they persist across power cycles.
        """
        return self._clib.fxSaveUTTToMemory(self.id)

    # -----
    # read_utts
    # -----
    @minimum_required_version("9.1.0")
    @requires_status("connected")
    def read_utts(self) -> List[int]:
        """
        UTTs are not sent as a part of regular communication, so here we
        first request that the device send the UTTs to Mn, and then we
        read them.
        """
        if self._clib.fxRequestUTT(self.id) != self._SUCCESS.value:
            raise RuntimeError("Error: could not request UTTs.")
        # We have to sleep for a bit in order to allow the device time
        # to fulfill the request
        sleep(0.25)

        numUtts = self.num_utts
        data = (c.c_int * numUtts)()
        retCode = self._clib.fxGetLastReceivedUTT(self.id, data, numUtts)

        if retCode != self._SUCCESS.value:
            raise RuntimeError("Error: could not read UTTs.")

        return [data[i] for i in range(numUtts)]

    # -----
    # gains
    # -----
    @property
    def gains(self) -> dict:
        """
        Returns the currently set gains.
        """
        return self._gains



import ctypes as c
import sys
from typing import List

import boto3
from botocore import UNSIGNED
from botocore.client import Config
from botocore.exceptions import EndpointConnectionError
import pendulum
from semantic_version import SimpleSpec
from semantic_version import Version
import yaml

from flexsea.utilities.aws import get_s3_objects
import flexsea.utilities.constants as fxc


# ============================================
#      validate_given_firmware_version
# ============================================
def validate_given_firmware_version(firmwareVersion: str, interactive: bool) -> Version:
    availableVersions = get_available_firmware_versions()

    # First we check if the user gave us a complete version string,
    # e.g., 7.2.0
    try:
        firmwareVersion = Version(firmwareVersion)
    except ValueError as err:
        # Check if the user gave us a partial version string, e.g., 7
        try:
            firmwareVersion = Version.coerce(firmwareVersion)
        except ValueError as err2:
            msg = f"Error: could not cast given version {firmwareVersion} to a valid "
            msg += "semantic version string. Please use the form X.Y.Z, where X, Y, "
            msg += "and Z are integers. For valid versions, please see: "
            msg += "flexsea.utilities.firmware.get_available_firmware_versions()"
            print(msg)
            raise err from err2

        # If the user did give us a partial version string, we now want
        # to find the most recent version that shares a major version
        # with what the user gave us. E.g., if the user gave us 7 and
        # the list of available versions contains 7.2.3 and 7.4.5, we
        # want to use 7.4.5
        latestVer = get_closest_version(firmwareVersion, availableVersions)
        if firmwareVersion != latestVer:
            msg = f"Warning: received version {firmwareVersion}, but found: "
            msg += f"{latestVer}, which is newer."
            print(msg)
            if interactive:
                userInput = input(f"Use {latestVer}? [y/n]")
                if userInput.lower() == "y":
                    firmwareVersion = latestVer
                else:
                    print("Aborting: no valid version selected.")
                    sys.exit(1)
            else:
                firmwareVersion = latestVer
    # If the user did give us a full version string, we make sure that
    # it's in the list of known versions
    else:
        try:
            assert str(firmwareVersion) in availableVersions
        except AssertionError as err:
            msg = f"Error: no library found for version: {firmwareVersion}. Use "
            msg += "flexsea.utilities.firmware.get_available_firmware_versions() "
            msg += "to see the available versions."
            raise ValueError(msg) from err

    return firmwareVersion


# ============================================
#       get_available_firmware_versions
# ============================================
def get_available_firmware_versions() -> List[str]:
    """
    To facilitiate offline use and firmware version validation, we
    cache a list of the currently available versions each time this
    function is run while online. If the user is not online, we load
    the available versions from the previously cached file and warn
    the user about how long it has been since their information was
    updated. If the file doesn't exist and we're not online, then
    the version information cannot be obtained, so we exit.
    """
    client = boto3.client(
        "s3", config=Config(signature_version=UNSIGNED), region_name="us-east-1"
    )

    try:
        objs = get_s3_objects(fxc.dephyPublicFilesBucket, client, prefix=fxc.libsDir)
    except EndpointConnectionError:
        print("Warning: unable to access S3 to obtain updated available versions.")
        try:
            with open(fxc.firmwareVersionCacheFile, "r", encoding="utf-8") as fd:
                data = yaml.safe_load(fd)
        except FileNotFoundError as err:
            msg = "Error: no firmware version cache file found. "
            msg += "Try connecting to the internet and running this function again."
            print(msg)
            raise err
        libs = data["versions"]
        days = (pendulum.today() - pendulum.parse(data["date"])).days
        if days > 7:
            print(f"Warning: using firmware version information from: {days} days ago.")
            print("To update, connect to the internet and re-run this function.")
    else:
        libs = set()
        for obj in objs:
            lib = obj.split("/")[1]
            libs.add(lib)
        libs = sorted(list(libs))

        with open(fxc.firmwareVersionCacheFile, "w", encoding="utf-8") as fd:
            yaml.safe_dump({"date": str(pendulum.today()), "versions": libs}, fd)
    finally:
        client.close()

    return libs


# ============================================
#             get_closest_version
# ============================================
def get_closest_version(version: Version, versionList: List[str]) -> Version:
    # https://tinyurl.com/3e4t6svb
    versionSpec = SimpleSpec(f"~={version.major}")
    closestVersion = versionSpec.select([Version(v) for v in versionList])

    if closestVersion is None:
        raise RuntimeError(f"Could not find version: {version}")

    return closestVersion


# ============================================
#                  Firmware
# ============================================
class Firmware(c.Structure):
    """
    Holds the integer values representing the firmware versions of
    each microcontroller as returned from the C library. These need to
    be decoded before they make sense.
    """

    _fields_ = [
        ("mn", c.c_uint32),
        ("ex", c.c_uint32),
        ("re", c.c_uint32),
        ("habs", c.c_uint32),
    ]


# ============================================
#              decode_firmware
# ============================================
def decode_firmware(val: int) -> str:
    """
    Returns decoded version number formatted as x.y.z
    """
    major: int = 0
    minor: int = 0
    bug: int = 0

    if val > 0:
        while val % 2 == 0:
            major += 1
            val = int(val / 2)

        while val % 3 == 0:
            minor += 1
            val = int(val / 3)

        while val % 5 == 0:
            bug += 1
            val = int(val / 5)

    return f"{major}.{minor}.{bug}"



from botocore.exceptions import EndpointConnectionError
import yaml
from semantic_version import Version

import flexsea.utilities.constants as fxc
from flexsea.utilities.aws import s3_download


# ============================================
#               get_device_spec
# ============================================
def get_device_spec(deviceName: str, firmwareVersion: Version) -> dict:
    deviceSpecFile = fxc.legacyDeviceSpecsPath.joinpath(
        str(firmwareVersion), f"{deviceName}.yaml"
    )

    if not deviceSpecFile.exists():
        deviceSpecFile.parent.mkdir(parents=True, exist_ok=True)
        deviceSpecObj = (
            f"{fxc.legacyDeviceSpecsDir}/{firmwareVersion}/{deviceSpecFile.name}"
        )
        try:
            s3_download(
                deviceSpecObj, fxc.dephyPublicFilesBucket, str(deviceSpecFile), None
            )
        except EndpointConnectionError as err:
            msg = "Error: could not connect to the internet to download the "
            msg += "necessary device spec file. Please connect to the internet and "
            msg += "try again."
            print(msg)
            raise err

    with open(deviceSpecFile, "r", encoding="utf-8") as fd:
        deviceSpec = yaml.safe_load(fd)

    return deviceSpec
